program estimate;

{$APPTYPE CONSOLE}

uses
  SysUtils,  StrUtils, Math;

type
  TStatKind = 0..3;
  TLevels = 0..62;

const
  MaxIterator: integer = 100000;

var
  iterator: integer;
  r, Divisor: int64;
  Has4kp3: boolean;
  Histogram: packed array [TStatKind, TLevels] of integer;

procedure InitHistogram;
var i, j: byte;
begin
  for i := Low(TStatKind) to High(TStatKind) do
    for j := Low(TLevels) to High(TLevels) do
      Histogram[i, j]:=0;
end;

function GSD(a, b: int64): int64;
var t: int64;
begin
  while (b>0) do begin
    t:=a mod b;
    a:=b;
    b:=t;
  end;
  Result:=a;
end;

function MaxDivisor(n: int64; var f: boolean): int64; // Lehman method
var
  Found: boolean;
  k, d, A, s, B, dx, dy: int64;
begin
  f:=false;
  s:=n;
  Result:=3;
  while (Result<=Power(s,1/3)) and not f do begin
    d:=trunc(s/Result);
    if s=d*Result then begin
      f:=Result mod 4=3;
      s:=d;
    end
    else inc(Result, 2);
  end;
  if (s=n) then begin
    Found:=false;
    k:=1;
    while not Found and (k<=Power(n,1/3)) do begin
      d:=0;
      while not Found and (d<=trunc(Power(n,1/6)/4/sqrt(k))+1) do begin
        A:=trunc(sqrt(4.0*k*n))+d;
        s:=A*A-4*k*n;
        if (s>0) then begin
          B:=trunc(sqrt(s));
          if (B*B=s) then begin
            if ((A*A) mod n = B*B) or (((A-B)*(A+B)) mod n = 0) then begin
              dx:=GSD(A-B, n);
              if (dx>1) and (dx<n) then begin
                dy:=trunc(n/dx);
                f:=(dx mod 4=3) or (dy mod 4=3);
                Result:=max(dx, dy);
                Found:=true;
              end;
            end;
          end;
        end;
        inc(d);
      end;
      inc(k);
    end;
    if not Found then Result:=n;
  end else Result:=s;
end;

function OldMaxDivisor(n: int64; var f: boolean): int64;
var
  d: int64;
begin
  f:=false;
  Result:=3;
  while (Result<=sqrt(n)) and not f do begin
    d:=trunc(n/Result);
    if n=d*Result then begin
      f:=Result mod 4=3;
      n:=d;
    end
    else inc(Result, 2);
  end;
  Result:=n;
end;

procedure TypeResults;
var i, j: byte;
begin
  for j := Low(TLevels) to High(TLevels) do begin
    write(DupeString(' ', 2-Length(IntToStr(j))), IntToStr(j));
    for i := Low(TStatKind) to High(TStatKind) do
      write(' | ', DupeString(' ', 6-Length(IntToStr(Histogram[i, j]))), Histogram[i, j]);
    writeln(' |');
  end;
  writeln(DupeString('=', 40));
end;

begin
  try
    Randomize;
    iterator:=0;
    InitHistogram;
    repeat
      r:=Trunc(Random*High(int64)/4)*4+1;
      Divisor:=MaxDivisor(r, Has4kp3);
      inc(Histogram[0][Trunc(ln(r)/ln(2))]); //All 4k+1 numbers
      if not Has4kp3 then begin
        inc(Histogram[1][Trunc(ln(r)/ln(2))]); //Has only 4k+1 divisors
        if Divisor = r then inc(Histogram[2][Trunc(ln(r)/ln(2))]) //Primes 4k+1
        else begin
          inc(Histogram[3][Trunc(ln(Divisor)/ln(2))]);
        end;
      end;
      inc(iterator);
      if iterator mod 10 = 0 then begin
        writeln(iterator);
        writeln(DupeString('=', 40));
        TypeResults;
      end;
    until iterator = MaxIterator;
    readln;
  except
    on E:Exception do
      Writeln(E.Classname, ': ', E.Message);
  end;
end.
